﻿namespace HeuristicLab.Problems.ProgramSynthesis.Base.Erc {
  using System;
  using System.Collections.Generic;
  using System.Linq;

  using Boolean;
  using Char;
  using Common;
  using Core;
  using Data;
  using Float;
  using FloatVector;
  using Integer;
  using IntegerVector;
  using Parameters;
  using Persistence.Default.CompositeSerializers.Storable;
  using String;
  using StringVector;

  [StorableClass]
  public class ErcOptions : ParameterizedNamedItem, IReadOnlyErcOptions {
    public const string ProbabilityParameterName = "Probability";
    public const string IntegerParameterName = "Integer";
    public const string FloatParameterName = "Float";
    public const string CharParameterName = "Char";
    public const string BooleanParameterName = "Boolean";
    public const string StringParameterName = "String";
    public const string NameParameterName = "Name";
    public const string IntegerVectorParameterName = "IntegerVector";
    public const string FloatVectorParameterName = "FloatVector";
    public const string StringVectorParameterName = "StringVector";

    [Storable]
    private readonly ReadOnlyCheckedItemCollection<IValueParameter> ercOptions;
    [Storable]
    private readonly NamedItemCollection<IValueParameter> ercOptionParameters;
    private IDictionary<IValueParameter, EventHandler<bool>> ercOptionEnabledChangedHandler;

    public ErcOptions() {
      Name = "ERC Options";

      Parameters.Add(new FixedValueParameter<PercentValue>(ProbabilityParameterName));

      ercOptionParameters = new NamedItemCollection<IValueParameter> {
        new ValueParameter<IntegerErcOptions>(IntegerParameterName),
        new ValueParameter<FloatErcOptions>(FloatParameterName),
        new ValueParameter<BooleanErcOptions>(BooleanParameterName),
        new ValueParameter<StringErcOptions>(StringParameterName),
        new ValueParameter<CharErcOptions>(CharParameterName),
        new ValueParameter<StringErcOptions>(NameParameterName),
        new ValueParameter<IntegerVectorErcOptions>(IntegerVectorParameterName),
        new ValueParameter<FloatVectorErcOptions>(FloatVectorParameterName),
        new ValueParameter<StringVectorErcOptions>(StringVectorParameterName)
      };

      InitEnabledChangedHandlerDict(ercOptionParameters);

      var optionsCollection = new CheckedItemCollection<IValueParameter>(ercOptionParameters);
      ercOptions = new ReadOnlyCheckedItemCollection<IValueParameter>(optionsCollection);
      Parameters.Add(new FixedValueParameter<ReadOnlyCheckedItemCollection<IValueParameter>>("Options", ercOptions));

      ercOptions.CheckedItemsChanged += ErcOptionsCheckedItemsChanged;

      DisableAllOptions();
    }

    [StorableConstructor]
    protected ErcOptions(bool deserializing) : base(deserializing) { }

    public ErcOptions(ErcOptions origin, Cloner cloner)
      : base(origin, cloner) {
      ercOptions = cloner.Clone(origin.ercOptions);
      ercOptionParameters = cloner.Clone(origin.ercOptionParameters);

      InitEnabledChangedHandlerDict(ercOptionParameters);
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitEnabledChangedHandlerDict(ercOptionParameters);
    }

    private void InitEnabledChangedHandlerDict(NamedItemCollection<IValueParameter> parameters) {
      ercOptionEnabledChangedHandler = parameters
        .ToDictionary<IValueParameter, IValueParameter, EventHandler<bool>>(
          param => param,
          param => (sender, state) => ercOptions.SetItemCheckedState(param, state));
    }

    private void ErcOptionsCheckedItemsChanged(object sender, Collections.CollectionItemsChangedEventArgs<IValueParameter> e) {
      foreach (var item in e.Items) {
        var ercItem = (IErcItem)item.Value;
        ercItem.IsEnabled = ercOptions.ItemChecked(item);
      }
    }

    public IValueParameter<PercentValue> ErcProbabilityParameter
    {
      get { return (IValueParameter<PercentValue>)Parameters[ProbabilityParameterName]; }
    }

    public double ErcProbability
    {
      get { return ErcProbabilityParameter.Value.Value; }
      set { ErcProbabilityParameter.Value.Value = value; }
    }

    IErcValueItem<int> IReadOnlyErcOptions.IntegerErcOptions { get { return IntegerErcOptions; } }
    IErcValueItem<double> IReadOnlyErcOptions.FloatErcOptions { get { return FloatErcOptions; } }
    IErcValueItem<bool> IReadOnlyErcOptions.BooleanErcOptions { get { return BooleanErcOptions; } }
    IErcValueItem<char> IReadOnlyErcOptions.CharErcOptions { get { return CharErcOptions; } }
    IErcValueItem<string> IReadOnlyErcOptions.StringErcOptions { get { return StringErcOptions; } }
    IErcValueItem<string> IReadOnlyErcOptions.NameErcOptions { get { return NameErcOptions; } }
    IErcValueItem<IReadOnlyList<int>> IReadOnlyErcOptions.IntegerVectorErcOptions { get { return IntegerVectorErcOptions; } }
    IErcValueItem<IReadOnlyList<double>> IReadOnlyErcOptions.FloatVectorErcOptions { get { return FloatVectorErcOptions; } }
    IErcValueItem<IReadOnlyList<string>> IReadOnlyErcOptions.StringVectorErcOptions { get { return StringVectorErcOptions; } }

    public IValueParameter<IntegerErcOptions> IntegerErcOptionsParameter
    {
      get { return (IValueParameter<IntegerErcOptions>)ercOptionParameters[IntegerParameterName]; }
    }

    public IntegerErcOptions IntegerErcOptions
    {
      get { return IntegerErcOptionsParameter.Value; }
      set { SetErcOptions(IntegerErcOptionsParameter, value); }
    }

    public IValueParameter<FloatErcOptions> FloatErcOptionsParameter
    {
      get { return (IValueParameter<FloatErcOptions>)ercOptionParameters[FloatParameterName]; }
    }

    public FloatErcOptions FloatErcOptions
    {
      get { return FloatErcOptionsParameter.Value; }
      set { SetErcOptions(FloatErcOptionsParameter, value); }
    }

    /// <summary>
    /// Integer represents ASCII value
    /// </summary>
    public IValueParameter<CharErcOptions> CharErcOptionsParameter
    {
      get { return (IValueParameter<CharErcOptions>)ercOptionParameters[CharParameterName]; }
    }

    public CharErcOptions CharErcOptions
    {
      get { return CharErcOptionsParameter.Value; }
      set { SetErcOptions(CharErcOptionsParameter, value); }
    }

    public IValueParameter<BooleanErcOptions> BooleanErcOptionsParameter
    {
      get { return (IValueParameter<BooleanErcOptions>)ercOptionParameters[BooleanParameterName]; }
    }

    public BooleanErcOptions BooleanErcOptions
    {
      get { return BooleanErcOptionsParameter.Value; }
      set { SetErcOptions(BooleanErcOptionsParameter, value); }
    }

    public IValueParameter<StringErcOptions> StringErcOptionsParameter
    {
      get { return (IValueParameter<StringErcOptions>)ercOptionParameters[StringParameterName]; }
    }

    public StringErcOptions StringErcOptions
    {
      get { return StringErcOptionsParameter.Value; }
      set { SetErcOptions(StringErcOptionsParameter, value); }
    }

    public IValueParameter<StringErcOptions> NameParameter
    {
      get { return (IValueParameter<StringErcOptions>)ercOptionParameters[NameParameterName]; }
    }

    public StringErcOptions NameErcOptions
    {
      get { return NameParameter.Value; }
      set { SetErcOptions(NameParameter, value); }
    }

    public IValueParameter<IntegerVectorErcOptions> IntegerVectorErcOptionsParameter
    {
      get { return (IValueParameter<IntegerVectorErcOptions>)ercOptionParameters[IntegerVectorParameterName]; }
    }

    public IntegerVectorErcOptions IntegerVectorErcOptions
    {
      get { return this.IntegerVectorErcOptionsParameter.Value; }
      set { SetErcOptions(this.IntegerVectorErcOptionsParameter, value); }
    }

    public IValueParameter<StringVectorErcOptions> StringVectorErcOptionsValueParameter
    {
      get { return (IValueParameter<StringVectorErcOptions>)ercOptionParameters[StringVectorParameterName]; }
    }

    public StringVectorErcOptions StringVectorErcOptions
    {
      get { return this.StringVectorErcOptionsValueParameter.Value; }
      set { SetErcOptions(this.StringVectorErcOptionsValueParameter, value); }
    }

    public IValueParameter<FloatVectorErcOptions> FloatVectorErcOptionsParameter
    {
      get { return (IValueParameter<FloatVectorErcOptions>)ercOptionParameters[FloatVectorParameterName]; }
    }

    public FloatVectorErcOptions FloatVectorErcOptions
    {
      get { return this.FloatVectorErcOptionsParameter.Value; }
      set { SetErcOptions(this.FloatVectorErcOptionsParameter, value); }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ErcOptions(this, cloner);
    }

    public void DisableAllOptions() {
      SetOptionsState(false);
    }

    public void EnableAllOptions() {
      SetOptionsState(true);
    }

    public void SetOptionsState(bool state) {
      foreach (var param in this.ercOptions) {
        this.ercOptions.SetItemCheckedState(param, state);
      }
    }

    private void SetErcOptions(IValueParameter param, IErcItem value) {
      if (ercOptionEnabledChangedHandler[param] != null && param.Value != null)
        ((IErcItem)param.Value).EnabledChanged -= ercOptionEnabledChangedHandler[param];

      param.Value = value;
      ercOptions.SetItemCheckedState(param, value != null ? value.IsEnabled : false);

      if (value != null)
        value.EnabledChanged += ercOptionEnabledChangedHandler[param];
    }
  }
}
