﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;
  using Attributes;
  using Common;
  using Interpreter;
  using Stack;

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the top two items are equal, or FALSE otherwise
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class EqualsExpression<T> : StatelessExpression {
    public bool Eval(IPushStack<T> stack, IPushStack<bool> booleanStack) {
      if (stack.Count < 2) return false;

      var first = stack[1];
      var second = stack.Top;
      stack.Remove(2);

      booleanStack.Push(first.Equals(second));

      return true;
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.=", StackTypes.Boolean)]
  public class IntegerEqualsExpression : EqualsExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.=", StackTypes.Boolean)]
  public class FloatEqualsExpression : EqualsExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2) return false;

      var first = interpreter.FloatStack[1];
      var second = interpreter.FloatStack.Top;
      interpreter.FloatStack.Remove(2);

      interpreter.BooleanStack.Push(first.IsAlmost(second));
      return true;
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.=", StackTypes.Boolean)]
  public class BooleanEqualsExpression : EqualsExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.=", StackTypes.Boolean)]
  public class NameEqualsExpression : EqualsExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.NameStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.=", StackTypes.Boolean)]
  public class ExecEqualsExpression : EqualsExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.ExecStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.=", StackTypes.Boolean)]
  public class CodeEqualsExpression : EqualsExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CodeStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.=", StackTypes.Boolean)]
  public class CharEqualsExpression : EqualsExpression<char> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CharStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.=", StackTypes.Boolean)]
  public class StringEqualsExpression : EqualsExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].=", StackTypes.Boolean)]
  public class IntegerVectorEqualsExpression : EqualsExpression<List<long>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].=", StackTypes.Boolean)]
  public class FloatVectorEqualsExpression : EqualsExpression<List<double>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].=", StackTypes.Boolean)]
  public class BooleanVectorEqualsExpression : EqualsExpression<List<bool>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].=", StackTypes.Boolean)]
  public class StringVectorEqualsExpression : EqualsExpression<List<string>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringVectorStack, interpreter.BooleanStack);
    }
  }
}