﻿using System;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;
  using System.Text;

  using Attributes;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Constants;

  using Interpreter;
  using Stack;

  [PushExpression(StackTypes.Print, "PRINT.NEWLINE")]
  public class PrintNewLineExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      interpreter.PrintStack.Push(Environment.NewLine);
      return true;
    }
  }

  public abstract class PrintExpression<T> : StatelessExpression {
    protected bool Eval(IInternalPushInterpreter interpreter, IPushStack<T> stack) {
      if (stack.IsEmpty)
        return false;

      var value = stack.Pop().ToString();
      interpreter.PrintStack.Push(value);
      return true;
    }

    protected bool EvalVector(IInternalPushInterpreter interpreter, IPushStack<List<T>> vectorStack) {
      if (vectorStack.IsEmpty)
        return false;

      var sb = new StringBuilder();
      var vector = vectorStack.Pop();

      sb.Append(PushEnvironment.VectorStartSymbol);

      if (vector.Count > 0) {
        sb.Append(vector[0]);

        for (var i = 1; i < vector.Count; i++) {
          sb.Append(PushEnvironment.VectorSeparatorSymbol);
          sb.Append(vector[i]);
        }
      }

      sb.Append(PushEnvironment.VectorEndSymbol);

      var value = sb.ToString();
      interpreter.PrintStack.Push(value);
      return true;
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTBOOLEAN", StackTypes.Boolean)]
  public class BooleanPrintExpression : PrintExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTCHAR", StackTypes.Char)]
  public class CharPrintExpression : PrintExpression<char> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.CharStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTEXEC", StackTypes.Exec)]
  public class ExecPrintExpression : PrintExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.ExecStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTFLOAT", StackTypes.Float)]
  public class FloatPrintExpression : PrintExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.FloatStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTINTEGER", StackTypes.Integer)]
  public class IntegerPrintExpression : PrintExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTSTRING", StackTypes.String)]
  public class StringPrintExpression : PrintExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.StringStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTINTEGERVECTOR", StackTypes.IntegerVector)]
  public class IntegerVectorPrintExpression : PrintExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return EvalVector(interpreter, interpreter.IntegerVectorStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTFLOATVECTOR", StackTypes.FloatVector)]
  public class FloatVectorPrintExpression : PrintExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return EvalVector(interpreter, interpreter.FloatVectorStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTSTRINGVECTOR", StackTypes.StringVector)]
  public class StringVectorPrintExpression : PrintExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return EvalVector(interpreter, interpreter.StringVectorStack);
    }
  }

  [PushExpression(StackTypes.Print, "PRINT.PRINTBOOLEANVECTOR", StackTypes.BooleanVector)]
  public class BooleanVectorPrintExpression : PrintExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return EvalVector(interpreter, interpreter.BooleanVectorStack);
    }
  }
}
